@@ -39,7 +39,7 @@ from .urls_patterns import URL_ADMIN_SERVER_INFO, URL_ADMIN_CLIENT_AUTHZ_RESOURC
     URL_ADMIN_GROUP_MEMBERS, URL_ADMIN_USER_STORAGE, URL_ADMIN_GROUP_PERMISSIONS, URL_ADMIN_IDPS, \
     URL_ADMIN_USER_CLIENT_ROLES_AVAILABLE, URL_ADMIN_USERS, URL_ADMIN_CLIENT_AUTHZ_RESOURCES_ID, URL_ADMIN_CLIENT_AUTHZ_POLICIES, \
     URL_ADMIN_CLIENT_AUTHZ_POLICIES_ID, URL_ADMIN_CLIENT_AUTHZ_PERMISSIONS, URL_ADMIN_CLIENT_AUTHZ_PERMISSIONS_ID, \
-    URL_ADMIN_USER_CLIENT_ROLES_COMPOSITE_ID
+    URL_ADMIN_USER_CLIENT_ROLES_COMPOSITE_ID, URL_ADMIN_GROUP_CLIENT_ROLES
 
 def serialize(obj):
     """JSON serializer for objects not serializable by default json code"""
@@ -200,6 +200,23 @@ class KeycloakAdmin:
         data_raw = self.connection.raw_get(URL_ADMIN_USERS_COUNT.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
+    def get_mapped_user_id(self, username, reference_db_id):
+        """
+        Get internal keycloak user id from username
+        This is required for further actions against this user.
+
+        UserRepresentation
+        http://www.keycloak.org/docs-api/3.3/rest-api/index.html#_userrepresentation
+
+        :param username: id in UserRepresentation
+
+        :return: user_id
+        """
+
+        query = {"username": username}
+        users = self.get_users(query)
+        return next((user["id"] for user in users if user["username"] == username and user["attributes"]["ReferenceId"][0] == reference_db_id), None)
+
     def get_user_id(self, username):
         """
         Get internal keycloak user id from username
@@ -213,7 +230,8 @@ class KeycloakAdmin:
         :return: user_id
         """
 
-        users = self.get_users(query={"username": username})
+        query = {"username": username}
+        users = self.get_users(query)
         return next((user["id"] for user in users if user["username"] == username), None)
 
     def get_user(self, user_id):
@@ -363,7 +381,28 @@ class KeycloakAdmin:
         data_raw = self.connection.raw_get(URL_ADMIN_SERVER_INFO)
         return raise_error_from_response(data_raw, KeycloakGetError)
 
-    def get_groups(self):
+    def get_mapped_group_id(self, groupname, reference_db_id):
+        """
+        Get internal keycloak group id from groupname
+        This is required for further actions against this group.
+
+        GroupRepresentation
+        http://www.keycloak.org/docs-api/3.3/rest-api/index.html#_grouprepresentation
+
+        :param groupname: id in GroupRepresentation
+
+        :return: group_id
+        """
+
+        query = {"name": groupname}
+        groups = self.get_groups(query)
+        group_id = next((group["id"] for group in groups if group["name"] == groupname), None)
+        if group_id is None:
+            return None
+        group = self.get_group(group_id)
+        return group_id if group["attributes"]["ReferenceId"][0] == reference_db_id else None
+
+    def get_groups(self, query=None):
         """
         Get groups belonging to the realm. Returns a list of groups belonging to the realm
 
@@ -373,7 +412,7 @@ class KeycloakAdmin:
         :return: array GroupRepresentation
         """
         params_path = {"realm-name": self.realm_name}
-        data_raw = self.connection.raw_get(URL_ADMIN_GROUPS.format(**params_path))
+        data_raw = self.connection.raw_get(URL_ADMIN_GROUPS.format(**params_path), **query)
         return raise_error_from_response(data_raw, KeycloakGetError)
 
     def get_group(self, group_id):
@@ -389,6 +428,23 @@ class KeycloakAdmin:
         data_raw = self.connection.raw_get(URL_ADMIN_GROUP.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
+    def get_group_id(self, groupname):
+        """
+        Get internal keycloak group id from groupname
+        This is required for further actions against this group.
+
+        GroupRepresentation
+        http://www.keycloak.org/docs-api/3.3/rest-api/index.html#_grouprepresentation
+
+        :param groupname: id in GroupRepresentation
+
+        :return: group_id
+        """
+
+        query = {}
+        groups = self.get_groups(query)
+        return next((group["id"] for group in groups if group["name"] == groupname), None)
+
     def get_subgroups(self, group, path):
         """
         Utility function to iterate through nested group structures
@@ -529,6 +585,48 @@ class KeycloakAdmin:
         data_raw = self.connection.raw_delete(URL_ADMIN_USER_GROUP.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
 
+    def assign_client_group_role(self, client_id, group_id, roles):
+        """
+        Assign a client role to a group
+
+        :param client_id: id of client containing role,
+        :param group_id: id of group
+        :param roles: roles list or role (use RoleRepresentation)
+        :return Keycloak server response
+        """
+
+        payload = roles if isinstance(roles, list) else [roles]
+        params_path = {"realm-name": self.realm_name, "id": group_id, "client-id": client_id}
+        data_raw = self.connection.raw_post(URL_ADMIN_GROUP_CLIENT_ROLES.format(**params_path),
+                                            data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+
+    def delete_client_group_roles(self, group_id, client_id, roles):
+        """
+        Delete client roles from a group.
+
+        :param client_id: id of client (not client-id)
+        :param group_id: id of group
+        :param client_id: id of client containing role,
+        :param roles: roles list or role to delete (use RoleRepresentation)
+        :return: Keycloak server response
+        """
+        payload = roles if isinstance(roles, list) else [roles]
+        params_path = {"realm-name": self.realm_name, "id": group_id, "client-id": client_id}
+        data_raw = self.connection.raw_delete(URL_ADMIN_GROUP_CLIENT_ROLES.format(**params_path),
+                                              data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+
+    def get_client_roles_of_group(self, group_id, client_id):
+        """
+        Get all client roles for a group.
+
+        :param client_id: id of client (not client-id)
+        :param group_id: id of group
+        :return: Keycloak server response (array RoleRepresentation)
+        """
+        return self._get_client_roles_of(URL_ADMIN_GROUP_CLIENT_ROLES, group_id, client_id)
+
     def delete_group(self, group_id):
         """
         Deletes a group in the Realm
@@ -866,7 +964,7 @@ class KeycloakAdmin:
 
     def delete_client_role(self, client_role_id, role_name):
         """
-        Create a client role
+        Delete a client role
 
         RoleRepresentation
         http://www.keycloak.org/docs-api/3.3/rest-api/index.html#_rolerepresentation
@@ -878,7 +976,7 @@ class KeycloakAdmin:
         data_raw = self.connection.raw_delete(URL_ADMIN_CLIENT_ROLE.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
 
-    def assign_client_role(self, client_id, user_id, roles):
+    def assign_client_user_role(self, client_id, user_id, roles):
         """
         Assign a client role to a user
 
@@ -902,7 +1000,7 @@ class KeycloakAdmin:
         :param user_id: id of user
         :return: Keycloak server response (array RoleRepresentation)
         """
-        return self._get_client_roles_of_user(URL_ADMIN_USER_CLIENT_ROLES, user_id, client_id)
+        return self._get_client_roles_of(URL_ADMIN_USER_CLIENT_ROLES, user_id, client_id)
 
     def get_available_client_roles_of_user(self, user_id, client_id):
         """
@@ -912,7 +1010,7 @@ class KeycloakAdmin:
         :param user_id: id of user
         :return: Keycloak server response (array RoleRepresentation)
         """
-        return self._get_client_roles_of_user(URL_ADMIN_USER_CLIENT_ROLES_AVAILABLE, user_id, client_id)
+        return self._get_client_roles_of(URL_ADMIN_USER_CLIENT_ROLES_AVAILABLE, user_id, client_id)
 
     def get_composite_client_roles_of_user(self, user_id, client_id):
         """
@@ -922,7 +1020,7 @@ class KeycloakAdmin:
         :param user_id: id of user
         :return: Keycloak server response (array RoleRepresentation)
         """
-        return self._get_client_roles_of_user(URL_ADMIN_USER_CLIENT_ROLES_COMPOSITE, user_id, client_id)
+        return self._get_client_roles_of(URL_ADMIN_USER_CLIENT_ROLES_COMPOSITE, user_id, client_id)
 
     def create_composite_client_role(self, role_id, payload, skip_exists=False):
         params_path = {"realm-name": self.realm_name, "role-id": role_id}
@@ -930,12 +1028,12 @@ class KeycloakAdmin:
                                             data=json.dumps(payload, default=serialize))
         return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204, skip_exists=skip_exists)
 
-    def _get_client_roles_of_user(self, client_level_role_mapping_url, user_id, client_id):
-        params_path = {"realm-name": self.realm_name, "id": user_id, "client-id": client_id}
+    def _get_client_roles_of(self, client_level_role_mapping_url, group_or_user_id, client_id):
+        params_path = {"realm-name": self.realm_name, "id": group_or_user_id, "client-id": client_id}
         data_raw = self.connection.raw_get(client_level_role_mapping_url.format(**params_path))
         return raise_error_from_response(data_raw, KeycloakGetError)
 
-    def delete_client_roles_of_user(self, user_id, client_id, roles):
+    def delete_client_user_roles(self, user_id, client_id, roles):
         """
         Delete client roles from a user.
 
